#include <complex>
#include <fstream>
#include <iostream>
#include <limits>
#include <vector>
#include <stdexcept>
#include <cstring>
#include <array>

#include "WolframLibrary.h"

DLLEXPORT mint WolframLibrary_getVersion(){
  return WolframLibraryVersion;
}

DLLEXPORT int WolframLibrary_initialize(WolframLibraryData){
  return LIBRARY_NO_ERROR;
}

DLLEXPORT void WolframLibrary_uninitialize(WolframLibraryData){
  return;
}

#include "QQbar_threshold/load_grid.hpp"
#include "QQbar_threshold/xsection.hpp"
#include "QQbar_threshold/energy_levels.hpp"
#include "QQbar_threshold/residues.hpp"
#include "QQbar_threshold/alpha_s.hpp"
#include "QQbar_threshold/scheme_conversion.hpp"
#include "QQbar_threshold/width.hpp"

using namespace QQbar_threshold::pert_order;

extern "C"{
  DLLEXPORT int
  load_grid(
      WolframLibraryData libData,
      mint Argc, MArgument *Args,
      MArgument Res
  );

  DLLEXPORT int
  load_nonresonant_grid(
      WolframLibraryData libData,
      mint Argc, MArgument *Args,
      MArgument Res
  );

  DLLEXPORT int
  Et_min(
      WolframLibraryData libData,
      mint Argc, MArgument *Args,
      MArgument Res
  );

  DLLEXPORT int
  Et_max(
      WolframLibraryData libData,
      mint Argc, MArgument *Args,
      MArgument Res
  );

  DLLEXPORT int
  Gammat_min(
      WolframLibraryData libData,
      mint Argc, MArgument *Args,
      MArgument Res
  );

  DLLEXPORT int
  Gammat_max(
      WolframLibraryData libData,
      mint Argc, MArgument *Args,
      MArgument Res
  );

  DLLEXPORT int
  alpha_s_bottom(
      WolframLibraryData libData,
      mint Argc, MArgument *Args,
      MArgument Res
  );

  DLLEXPORT int
  alpha_s_top(
      WolframLibraryData libData,
      mint Argc, MArgument *Args,
      MArgument Res
  );

  DLLEXPORT int
  bbbar_xsection(
      WolframLibraryData libData,
      mint Argc, MArgument *Args,
      MArgument Res
  );

  DLLEXPORT int
  ttbar_xsection(
      WolframLibraryData libData,
      mint Argc, MArgument *Args,
      MArgument Res
  );

  DLLEXPORT int
  bbbar_R_ratio(
      WolframLibraryData libData,
      mint Argc, MArgument *Args,
      MArgument Res
  );

  DLLEXPORT int
  ttbar_R_ratio(
      WolframLibraryData libData,
      mint Argc, MArgument *Args,
      MArgument Res
  );

  DLLEXPORT int
  bbbar_energy_level(
      WolframLibraryData libData,
      mint Argc, MArgument *Args,
      MArgument Res
  );

  DLLEXPORT int
  ttbar_energy_level(
      WolframLibraryData libData,
      mint Argc, MArgument *Args,
      MArgument Res
  );

  DLLEXPORT int
  bbbar_residue(
      WolframLibraryData libData,
      mint Argc, MArgument *Args,
      MArgument Res
  );

  DLLEXPORT int
  ttbar_residue(
      WolframLibraryData libData,
      mint Argc, MArgument *Args,
      MArgument Res
  );

  DLLEXPORT int
  is_shifted_mass(
      WolframLibraryData libData,
      mint Argc, MArgument *Args,
      MArgument Res
  );

  DLLEXPORT int
  bottom_pole_mass(
      WolframLibraryData libData,
      mint Argc, MArgument *Args,
      MArgument Res
  );

  DLLEXPORT int
  top_pole_mass(
      WolframLibraryData libData,
      mint Argc, MArgument *Args,
      MArgument Res
  );

  DLLEXPORT int
  bbbar_threshold(
      WolframLibraryData libData,
      mint Argc, MArgument *Args,
      MArgument Res
  );

  DLLEXPORT int
  ttbar_threshold(
      WolframLibraryData libData,
      mint Argc, MArgument *Args,
      MArgument Res
  );

  DLLEXPORT int
  top_internal_settings_3(
      WolframLibraryData libData,
      mint Argc, MArgument *Args,
      MArgument Res
  );

  DLLEXPORT int
  bottom_internal_settings_3(
      WolframLibraryData libData,
      mint Argc, MArgument *Args,
      MArgument Res
  );

  DLLEXPORT int
  bottom_internal_settings_4(
      WolframLibraryData libData,
      mint Argc, MArgument *Args,
      MArgument Res
  );

  DLLEXPORT int
  top_internal_settings_6(
      WolframLibraryData libData,
      mint Argc, MArgument *Args,
      MArgument Res
  );

  DLLEXPORT int
  top_width(
      WolframLibraryData libData,
      mint Argc, MArgument *Args,
      MArgument Res
  );
}

DLLEXPORT int
load_grid(
    WolframLibraryData libData,
    mint Argc, MArgument *Args,
    MArgument /* Res */
){
  if(Argc != 1){
    return LIBRARY_FUNCTION_ERROR;
  }
  const char * grid_name = *(Args[0].utf8string);
  try{
    QQbar_threshold::load_grid(grid_name);
  }
  catch(std::exception & fail) {
    libData->Message(fail.what());
    return LIBRARY_FUNCTION_ERROR;
  }
  return LIBRARY_NO_ERROR;
}

DLLEXPORT int
load_nonresonant_grid(
    WolframLibraryData libData,
    mint Argc, MArgument *Args,
    MArgument /* Res */
){
  if(Argc != 1){
    return LIBRARY_FUNCTION_ERROR;
  }
  const char * grid_name = *(Args[0].utf8string);
  try{
    QQbar_threshold::load_nonresonant_grid(grid_name);
  }
  catch(std::exception & fail) {
    libData->Message(fail.what());
    return LIBRARY_FUNCTION_ERROR;
  }
  return LIBRARY_NO_ERROR;
}

DLLEXPORT int
Et_min(
    WolframLibraryData libData,
    mint /* Argc */, MArgument * /* Args */,
    MArgument Res
){
  try{
    MArgument_setReal(Res, QQbar_threshold::grid_range().Et_min);
  }
  catch(std::exception & fail) {
    libData->Message(fail.what());
    return LIBRARY_FUNCTION_ERROR;
  }
  return LIBRARY_NO_ERROR;
}

DLLEXPORT int
Et_max(
    WolframLibraryData libData,
    mint /* Argc */, MArgument * /* Args */,
    MArgument Res
){
  try{
    MArgument_setReal(Res, QQbar_threshold::grid_range().Et_max);
  }
  catch(std::exception & fail) {
    libData->Message(fail.what());
    return LIBRARY_FUNCTION_ERROR;
  }
  return LIBRARY_NO_ERROR;
}

DLLEXPORT int
Gammat_min(
    WolframLibraryData libData,
    mint /* Argc */, MArgument * /* Args */,
    MArgument Res
){
  try{
    MArgument_setReal(Res, QQbar_threshold::grid_range().Gammat_min);
  }
  catch(std::exception & fail) {
    libData->Message(fail.what());
    return LIBRARY_FUNCTION_ERROR;
  }
  return LIBRARY_NO_ERROR;
}

DLLEXPORT int
Gammat_max(
    WolframLibraryData libData,
    mint /* Argc */, MArgument * /* Args */,
    MArgument Res
){
  try{
    MArgument_setReal(Res, QQbar_threshold::grid_range().Gammat_max);
  }
  catch(std::exception & fail) {
    libData->Message(fail.what());
    return LIBRARY_FUNCTION_ERROR;
  }
  return LIBRARY_NO_ERROR;
}

DLLEXPORT int
alpha_s_bottom(
    WolframLibraryData libData,
    mint Argc, MArgument *Args,
    MArgument Res
){
  if(Argc != 3){
    return LIBRARY_FUNCTION_ERROR;
  }
  size_t cur_arg = 0;
  const double mu = MArgument_getReal(Args[cur_arg++]);
  double mu0 = MArgument_getReal(Args[cur_arg++]);
  if(mu0 == 0.0) mu0 = QQbar_threshold::mZ;
  const double alpha_s = MArgument_getReal(Args[cur_arg++]);
  try{
    MArgument_setReal(
        Res,
        QQbar_threshold::alpha_s_bottom(mu, {mu0, alpha_s})
    );
  }
  catch(std::exception & fail) {
    libData->Message(fail.what());
    return LIBRARY_FUNCTION_ERROR;
  }
  return LIBRARY_NO_ERROR;
}

DLLEXPORT int
alpha_s_top(
    WolframLibraryData libData,
    mint Argc, MArgument *Args,
    MArgument Res
){
  if(Argc != 3){
    return LIBRARY_FUNCTION_ERROR;
  }
  size_t cur_arg = 0;
  const double mu = MArgument_getReal(Args[cur_arg++]);
  double mu0 = MArgument_getReal(Args[cur_arg++]);
  if(mu0 == 0.0) mu0 = QQbar_threshold::mZ;
  const double alpha_s = MArgument_getReal(Args[cur_arg++]);
  try{
    MArgument_setReal(
        Res,
        QQbar_threshold::alpha_s_top(mu, {mu0, alpha_s})
    );
  }
  catch(std::exception & fail) {
    libData->Message(fail.what());
    return LIBRARY_FUNCTION_ERROR;
  }
  return LIBRARY_NO_ERROR;
}

namespace{
  constexpr int num_opts = 42;

  QQbar_threshold::options get_opt(
      WolframLibraryData /*libData*/,
      MArgument *Args
  ){
    QQbar_threshold::options opt;
    size_t cur_arg = 0;
    opt.mass_scheme.scale = MArgument_getReal(Args[cur_arg++]);
    opt.resum_poles = MArgument_getInteger(Args[cur_arg++]);
    opt.beyond_QCD = static_cast<QQbar_threshold::SM>(
        MArgument_getInteger(Args[cur_arg++])
    );
    opt.mass_scheme.name = static_cast<QQbar_threshold::scheme_name>(
        MArgument_getInteger(Args[cur_arg++])
    );
    opt.production = static_cast<QQbar_threshold::production_channel>(
        MArgument_getInteger(Args[cur_arg++])
    );
    opt.resonant_only = MArgument_getBoolean(Args[cur_arg++]);
    opt.expand_s = MArgument_getBoolean(Args[cur_arg++]);
    opt.double_light_insertion = MArgument_getBoolean(Args[cur_arg++]);
    opt.m_Higgs = MArgument_getReal(Args[cur_arg++]);
    opt.Yukawa_factor = MArgument_getReal(Args[cur_arg++]);
    opt.invariant_mass_cut = MArgument_getReal(Args[cur_arg++]);
    //default is signalled by a negative value
    if(opt.invariant_mass_cut < 0){
      opt.invariant_mass_cut = std::numeric_limits<double>::quiet_NaN();
    }
    opt.ml = MArgument_getReal(Args[cur_arg++]);
    opt.r4 = MArgument_getReal(Args[cur_arg++]);
    opt.alpha = MArgument_getReal(Args[cur_arg++]);
    opt.mu_alpha = MArgument_getReal(Args[cur_arg++]);
    opt.alpha_s_mu0.mu0 = MArgument_getReal(Args[cur_arg++]);
    opt.alpha_s_mu0.alpha_s = MArgument_getReal(Args[cur_arg++]);
    for(size_t i = 0; i < opt.contributions.v_Coulomb.size(); ++ i){
      opt.contributions.v_Coulomb[i] = MArgument_getReal(Args[cur_arg++]);
    }
    for(size_t i = 0; i < opt.contributions.v_delta.size(); ++ i){
      opt.contributions.v_delta[i] = MArgument_getReal(Args[cur_arg++]);
    }
    for(size_t i = 0; i < opt.contributions.v_r2inv.size(); ++ i){
      opt.contributions.v_r2inv[i] = MArgument_getReal(Args[cur_arg++]);
    }
    for(size_t i = 0; i < opt.contributions.v_p2.size(); ++ i){
      opt.contributions.v_p2[i] = MArgument_getReal(Args[cur_arg++]);
    }
    for(size_t i = 0; i < opt.contributions.v_kinetic.size(); ++ i){
      opt.contributions.v_kinetic[i] = MArgument_getReal(Args[cur_arg++]);
    }
    for(size_t i = 0; i < opt.contributions.ultrasoft.size(); ++ i){
      opt.contributions.ultrasoft[i] = MArgument_getReal(Args[cur_arg++]);
    }
    for(size_t i = 0; i < opt.contributions.v_Higgs.size(); ++ i){
      opt.contributions.v_Higgs[i] = MArgument_getReal(Args[cur_arg++]);
    }
    for(size_t i = 0; i < opt.contributions.v_QED_Coulomb.size(); ++ i){
      opt.contributions.v_QED_Coulomb[i] = MArgument_getReal(Args[cur_arg++]);
    }
    for(size_t i = 0; i < opt.contributions.cv.size(); ++ i){
      opt.contributions.cv[i] = MArgument_getReal(Args[cur_arg++]);
    }
    for(size_t i = 0; i < opt.contributions.cv_Higgs.size(); ++ i){
      opt.contributions.cv_Higgs[i] = MArgument_getReal(Args[cur_arg++]);
    }
    opt.contributions.Cv_QED[0] = MArgument_getReal(Args[cur_arg++]);
    opt.contributions.Cv_WZ[0] = MArgument_getReal(Args[cur_arg++]);
    opt.contributions.Ca_QED[0] = MArgument_getReal(Args[cur_arg++]);
    opt.contributions.Ca_WZ[0] = MArgument_getReal(Args[cur_arg++]);
    for(size_t i = 0; i < opt.contributions.dv.size(); ++ i){
      opt.contributions.dv[i] = MArgument_getReal(Args[cur_arg++]);
    }
    for(size_t i = 0; i < opt.contributions.ca.size(); ++ i){
      opt.contributions.ca[i] = MArgument_getReal(Args[cur_arg++]);
    }
    return opt;
  }
}

DLLEXPORT int
bbbar_xsection(
    WolframLibraryData libData,
    mint Argc, MArgument *Args,
    MArgument Res
){
  if(Argc != 4 + num_opts){
    return LIBRARY_FUNCTION_ERROR;
  }
  size_t cur_arg = 0;

  double sqrt_s = MArgument_getReal(Args[cur_arg++]);
  double mu = MArgument_getReal(Args[cur_arg++]);
  double m = MArgument_getReal(Args[cur_arg++]);
  order o = static_cast<order>(MArgument_getInteger(Args[cur_arg++]));

  try{
    MArgument_setReal(
	Res,
	bbbar_xsection(sqrt_s, mu, m, o, get_opt(libData, Args + cur_arg))
    );
  }
  catch(std::exception & fail){
    libData->Message(fail.what());
    return LIBRARY_FUNCTION_ERROR;
  }
  return LIBRARY_NO_ERROR;
}

DLLEXPORT int
ttbar_xsection(
    WolframLibraryData libData,
    mint Argc, MArgument *Args,
    MArgument Res
){
  if(Argc != 6 + num_opts){
    return LIBRARY_FUNCTION_ERROR;
  }
  size_t cur_arg = 0;

  double sqrt_s = MArgument_getReal(Args[cur_arg++]);
  double mu_soft = MArgument_getReal(Args[cur_arg++]);
  double mu_weak = MArgument_getReal(Args[cur_arg++]);
  double m = MArgument_getReal(Args[cur_arg++]);
  double width = MArgument_getReal(Args[cur_arg++]);
  order o = static_cast<order>(MArgument_getInteger(Args[cur_arg++]));

  try{
    MArgument_setReal(
	Res,
	ttbar_xsection(
	    sqrt_s, {mu_soft, mu_weak}, {m, width}, o,
	    get_opt(libData, Args + cur_arg)
	)
    );
  }
  catch(std::exception & fail){
    libData->Message(fail.what());
    return LIBRARY_FUNCTION_ERROR;
  }
  return LIBRARY_NO_ERROR;
}

DLLEXPORT int
bbbar_R_ratio(
    WolframLibraryData libData,
    mint Argc, MArgument *Args,
    MArgument Res
){
  if(Argc != 4 + num_opts){
    return LIBRARY_FUNCTION_ERROR;
  }
  size_t cur_arg = 0;

  double sqrt_s = MArgument_getReal(Args[cur_arg++]);
  double mu = MArgument_getReal(Args[cur_arg++]);
  double m = MArgument_getReal(Args[cur_arg++]);
  order o = static_cast<order>(MArgument_getInteger(Args[cur_arg++]));

  try{
    MArgument_setReal(
	Res,
	bbbar_R_ratio(sqrt_s, mu, m, o, get_opt(libData, Args + cur_arg))
    );
  }
  catch(std::exception & fail){
    libData->Message(fail.what());
    return LIBRARY_FUNCTION_ERROR;
  }
  return LIBRARY_NO_ERROR;
}

DLLEXPORT int
ttbar_R_ratio(
    WolframLibraryData libData,
    mint Argc, MArgument *Args,
    MArgument Res
){
  if(Argc != 6 + num_opts){
    return LIBRARY_FUNCTION_ERROR;
  }
  size_t cur_arg = 0;

  double sqrt_s = MArgument_getReal(Args[cur_arg++]);
  double mu_soft = MArgument_getReal(Args[cur_arg++]);
  double mu_weak = MArgument_getReal(Args[cur_arg++]);
  double m = MArgument_getReal(Args[cur_arg++]);
  double width = MArgument_getReal(Args[cur_arg++]);
  order o = static_cast<order>(MArgument_getInteger(Args[cur_arg++]));

  try{
    MArgument_setReal(
	Res,
	ttbar_R_ratio(
	    sqrt_s, {mu_soft, mu_weak}, {m, width}, o,
	    get_opt(libData, Args + cur_arg)
	)
    );
  }
  catch(std::exception & fail){
    libData->Message(fail.what());
    return LIBRARY_FUNCTION_ERROR;
  }
  return LIBRARY_NO_ERROR;
}

DLLEXPORT int
bbbar_energy_level(
    WolframLibraryData libData,
    mint Argc, MArgument *Args,
    MArgument Res
){
  if(Argc != 4 + num_opts){
    return LIBRARY_FUNCTION_ERROR;
  }
  size_t cur_arg = 0;

  int n = MArgument_getInteger(Args[cur_arg++]);
  double mu = MArgument_getReal(Args[cur_arg++]);
  double m = MArgument_getReal(Args[cur_arg++]);
  order o = static_cast<order>(MArgument_getInteger(Args[cur_arg++]));

  try{
    MArgument_setReal(
	Res,
	bbbar_energy_level(n, mu, m, o, get_opt(libData, Args + cur_arg))
    );
  }
  catch(std::exception & fail){
    libData->Message(fail.what());
    return LIBRARY_FUNCTION_ERROR;
  }
  return LIBRARY_NO_ERROR;
}

DLLEXPORT int
ttbar_energy_level(
    WolframLibraryData libData,
    mint Argc, MArgument *Args,
    MArgument Res
){
  if(Argc != 4 + num_opts){
    return LIBRARY_FUNCTION_ERROR;
  }
  size_t cur_arg = 0;

  int n = MArgument_getInteger(Args[cur_arg++]);
  double mu = MArgument_getReal(Args[cur_arg++]);
  double m = MArgument_getReal(Args[cur_arg++]);
  order o = static_cast<order>(MArgument_getInteger(Args[cur_arg++]));

  try{
    MArgument_setReal(
	Res,
	ttbar_energy_level(n, mu, m, o, get_opt(libData, Args + cur_arg))
    );
  }
  catch(std::exception & fail){
    libData->Message(fail.what());
    return LIBRARY_FUNCTION_ERROR;
  }
  return LIBRARY_NO_ERROR;
}

DLLEXPORT int
bbbar_residue(
    WolframLibraryData libData,
    mint Argc, MArgument *Args,
    MArgument Res
){
  if(Argc != 4 + num_opts){
    return LIBRARY_FUNCTION_ERROR;
  }
  size_t cur_arg = 0;

  int n = MArgument_getInteger(Args[cur_arg++]);
  double mu = MArgument_getReal(Args[cur_arg++]);
  double m = MArgument_getReal(Args[cur_arg++]);
  order o = static_cast<order>(MArgument_getInteger(Args[cur_arg++]));

  try{
    MArgument_setReal(
	Res,
	bbbar_residue(n, mu, m, o, get_opt(libData, Args + cur_arg))
    );
  }
  catch(std::exception & fail){
    libData->Message(fail.what());
    return LIBRARY_FUNCTION_ERROR;
  }
  return LIBRARY_NO_ERROR;
}

DLLEXPORT int
ttbar_residue(
    WolframLibraryData libData,
    mint Argc, MArgument *Args,
    MArgument Res
){
  if(Argc != 4 + num_opts){
    return LIBRARY_FUNCTION_ERROR;
  }
  size_t cur_arg = 0;

  int n = MArgument_getInteger(Args[cur_arg++]);
  double mu = MArgument_getReal(Args[cur_arg++]);
  double m = MArgument_getReal(Args[cur_arg++]);
  order o = static_cast<order>(MArgument_getInteger(Args[cur_arg++]));

  try{
    MArgument_setReal(
	Res,
	ttbar_residue(n, mu, m, o, get_opt(libData, Args + cur_arg))
    );
  }
  catch(std::exception & fail){
    libData->Message(fail.what());
    return LIBRARY_FUNCTION_ERROR;
  }
  return LIBRARY_NO_ERROR;
}

DLLEXPORT int
is_shifted_mass(
    WolframLibraryData libData,
    mint Argc, MArgument *Args,
    MArgument Res
){
  if(Argc != 1){
    return LIBRARY_FUNCTION_ERROR;
  }
  try{
    QQbar_threshold::scheme_name const mass_scheme =
      static_cast<QQbar_threshold::scheme_name>(
	MArgument_getInteger(Args[0])
    );
    MArgument_setBoolean(Res,is_shifted_mass(mass_scheme));
  }
  catch(std::exception & fail){
    libData->Message(fail.what());
    return LIBRARY_FUNCTION_ERROR;
  }
  return LIBRARY_NO_ERROR;
}

DLLEXPORT int
bottom_pole_mass(
    WolframLibraryData libData,
    mint Argc, MArgument *Args,
    MArgument Res
){
  if(Argc != 3 + num_opts){
    return LIBRARY_FUNCTION_ERROR;
  }
  size_t cur_arg = 0;

  double mu = MArgument_getReal(Args[cur_arg++]);
  double m = MArgument_getReal(Args[cur_arg++]);
  order o = static_cast<order>(MArgument_getInteger(Args[cur_arg++]));

  try{
    MArgument_setReal(
	Res,
	QQbar_threshold::bottom_pole_mass(
	    mu, m, o, get_opt(libData, Args + cur_arg)
	)
    );
  }
  catch(std::exception & fail){
    libData->Message(fail.what());;
    return LIBRARY_FUNCTION_ERROR;
  }
  return LIBRARY_NO_ERROR;
}

DLLEXPORT int
top_pole_mass(
    WolframLibraryData libData,
    mint Argc, MArgument *Args,
    MArgument Res
){
  if(Argc != 3 + num_opts){
    return LIBRARY_FUNCTION_ERROR;
  }
  size_t cur_arg = 0;

  double mu = MArgument_getReal(Args[cur_arg++]);
  double m = MArgument_getReal(Args[cur_arg++]);
  order o = static_cast<order>(MArgument_getInteger(Args[cur_arg++]));

  try{
    MArgument_setReal(
	Res,
	top_pole_mass(mu, m, o, get_opt(libData, Args + cur_arg))
    );
  }
  catch(std::exception & fail){
    libData->Message(fail.what());;
    return LIBRARY_FUNCTION_ERROR;
  }
  return LIBRARY_NO_ERROR;
}

DLLEXPORT int
bbbar_threshold(
    WolframLibraryData libData,
    mint Argc, MArgument *Args,
    MArgument Res
){
  if(Argc != 3 + num_opts){
    return LIBRARY_FUNCTION_ERROR;
  }
  size_t cur_arg = 0;

  double mu = MArgument_getReal(Args[cur_arg++]);
  double m = MArgument_getReal(Args[cur_arg++]);
  order o = static_cast<order>(MArgument_getInteger(Args[cur_arg++]));

  try{
    MArgument_setReal(
	Res,
	bbbar_threshold(mu, m, o, get_opt(libData, Args + cur_arg))
    );
  }
  catch(std::exception & fail){
    libData->Message(fail.what());;
    return LIBRARY_FUNCTION_ERROR;
  }
  return LIBRARY_NO_ERROR;
}

DLLEXPORT int
ttbar_threshold(
    WolframLibraryData libData,
    mint Argc, MArgument *Args,
    MArgument Res
){
  if(Argc != 3 + num_opts){
    return LIBRARY_FUNCTION_ERROR;
  }
  size_t cur_arg = 0;

  double mu = MArgument_getReal(Args[cur_arg++]);
  double m = MArgument_getReal(Args[cur_arg++]);
  order o = static_cast<order>(MArgument_getInteger(Args[cur_arg++]));

  try{
    MArgument_setReal(
	Res,
	ttbar_threshold(mu, m, o, get_opt(libData, Args + cur_arg))
    );
  }
  catch(std::exception & fail){
    libData->Message(fail.what());;
    return LIBRARY_FUNCTION_ERROR;
  }
  return LIBRARY_NO_ERROR;
}

namespace{
  /*
   * helper functions for bottom_internal_settings_{3,4}
   * and top_internal_settings_{3,5}
   */

  void set_Mathematica_String(MArgument Res, std::string const & str){
    /*
     * LibraryLink wants to convert our std::string to a UTF8String
     * and needs a char*. std::string::c_str() is a char const * and applying
     * a const_cast would lead to undefined behaviour. So we perform a copy.
     */
    static std::vector<char> result;
    result.resize(str.size() + 1);
    std::strncpy(result.data(), str.c_str(), result.size());
    assert(result.back() == '\0');
    MArgument_setUTF8String(Res, result.data());
  }

  template<typename... T>
    int bottom_internal_settings(
	WolframLibraryData libData,
	MArgument *Opts,
	MArgument Res,
	T&&... args
    ){
    try{
      const std::string settings = bottom_internal_settings(
	  std::forward<T>(args)...,
	  get_opt(libData, Opts)
      );
      set_Mathematica_String(Res, settings);
    }
    catch(std::exception & fail){
      libData->Message(fail.what());
      return LIBRARY_FUNCTION_ERROR;
    }
    return LIBRARY_NO_ERROR;
  }

  template<typename... T>
    int top_internal_settings(
	WolframLibraryData libData,
	MArgument *Opts,
	MArgument Res,
	T&&... args
    ){
    try{
      const std::string settings = top_internal_settings(
	  std::forward<T>(args)...,
	  get_opt(libData, Opts)
      );
      set_Mathematica_String(Res, settings);
    }
    catch(std::exception & fail){
      libData->Message(fail.what());
      return LIBRARY_FUNCTION_ERROR;
    }
    return LIBRARY_NO_ERROR;
  }

}

DLLEXPORT int
bottom_internal_settings_3(
    WolframLibraryData libData,
    mint Argc, MArgument *Args,
    MArgument Res
){
  if(Argc != 3 + num_opts){
    return LIBRARY_FUNCTION_ERROR;
  }
  size_t cur_arg = 0;

  double mu = MArgument_getReal(Args[cur_arg++]);
  double m = MArgument_getReal(Args[cur_arg++]);
  order o = static_cast<order>(MArgument_getInteger(Args[cur_arg++]));
  return bottom_internal_settings(
      libData, Args + cur_arg, Res,
      mu, m, o
  );
}

DLLEXPORT int
top_internal_settings_3(
    WolframLibraryData libData,
    mint Argc, MArgument *Args,
    MArgument Res
){
  if(Argc != 3 + num_opts){
    return LIBRARY_FUNCTION_ERROR;
  }
  size_t cur_arg = 0;

  double mu = MArgument_getReal(Args[cur_arg++]);
  double m = MArgument_getReal(Args[cur_arg++]);
  order o = static_cast<order>(MArgument_getInteger(Args[cur_arg++]));
  return top_internal_settings(
      libData, Args + cur_arg, Res,
      mu, m, o
  );
}

DLLEXPORT int
bottom_internal_settings_4(
    WolframLibraryData libData,
    mint Argc, MArgument *Args,
    MArgument Res
){
  if(Argc != 4 + num_opts){
    return LIBRARY_FUNCTION_ERROR;
  }
  size_t cur_arg = 0;

  double sqrt_s = MArgument_getReal(Args[cur_arg++]);
  double mu = MArgument_getReal(Args[cur_arg++]);
  double m = MArgument_getReal(Args[cur_arg++]);
  order o = static_cast<order>(MArgument_getInteger(Args[cur_arg++]));
  return bottom_internal_settings(
      libData, Args + cur_arg, Res,
      sqrt_s, mu, m, o
  );
}

DLLEXPORT int
top_internal_settings_6(
    WolframLibraryData libData,
    mint Argc, MArgument *Args,
    MArgument Res
){
  if(Argc != 6 + num_opts){
    return LIBRARY_FUNCTION_ERROR;
  }
  size_t cur_arg = 0;

  double sqrt_s = MArgument_getReal(Args[cur_arg++]);
  double mu_soft = MArgument_getReal(Args[cur_arg++]);
  double mu_weak = MArgument_getReal(Args[cur_arg++]);
  double m = MArgument_getReal(Args[cur_arg++]);
  double width = MArgument_getReal(Args[cur_arg++]);
  order o = static_cast<order>(MArgument_getInteger(Args[cur_arg++]));
  return top_internal_settings(
      libData, Args + cur_arg, Res,
      sqrt_s, QQbar_threshold::scales{mu_soft, mu_weak},
      QQbar_threshold::top_properties{m, width}, o
  );
}

DLLEXPORT int
top_width(
    WolframLibraryData libData,
    mint Argc, MArgument *Args,
    MArgument Res
){
  if(Argc != 3 + num_opts){
    return LIBRARY_FUNCTION_ERROR;
  }
  size_t cur_arg = 0;

  double mu = MArgument_getReal(Args[cur_arg++]);
  double m = MArgument_getReal(Args[cur_arg++]);
  order o = static_cast<order>(MArgument_getInteger(Args[cur_arg++]));
  try{
    MArgument_setReal(
	Res,
	QQbar_threshold::top_width(
	    mu, m, o, get_opt(libData, Args + cur_arg)
	)
    );
  }
  catch(std::exception & fail){
    libData->Message(fail.what());;
    return LIBRARY_FUNCTION_ERROR;
  }
  return LIBRARY_NO_ERROR;
}
