/**
 * @file
 * @brief Implementation of dilogarithm
 * @author  Andreas Maier <a.maier@tum.de>
 * @version 1.0.4
 *
 * @section LICENSE
 *
 * Copyright 2014, 2015 Andreas Maier
 *
 * This file is part of the QQbar_threshold library.
 *
 * The QQbar_threshold library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 3 of
 * the License, or (at your option) any later version.
 *
 * The QQbar_threshold library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the QQbar_threshold library. If not, see
 * <http://www.gnu.org/licenses/>.
 *
 */
#pragma once

#include <complex>

#include "gsl/gsl_sf.h"

namespace dilog{
  //! Dilogarithm for real argument
  inline double Li2(double x){
    return gsl_sf_dilog(x);
  }

  //! Dilogarithm for complex argument
  inline std::complex<double> Li2(std::complex<double> x){
    gsl_sf_result re,im;
    gsl_sf_complex_dilog_e(abs(x), arg(x), &re, &im);
    return {re.val, im.val};
  }
}
