/**
 * @file
 * @brief Implementation of default strong coupling
 *
 * @see alpha_s.hpp
 *
 * @author  Andreas Maier <a.maier@tum.de>
 * @version 1.0.4
 *
 * @section LICENSE
 *
 * Copyright 2014, 2015 Andreas Maier
 *
 * This file is part of the QQbar_threshold library.
 *
 * The QQbar_threshold library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 3 of
 * the License, or (at your option) any later version.
 *
 * The QQbar_threshold library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the QQbar_threshold library. If not, see
 * <http://www.gnu.org/licenses/>.
 *
 */
#include "alpha_s.hpp"
#include "constants.hpp"
#include "parameters.hpp"

#include "QCD/alpha_s.hpp"

namespace QQbar_threshold{

  double alpha_s_top(double mu, double alpha_s_mZ){
    return alpha_s_top(mu, {mZ, alpha_s_mZ});
  }

  double alpha_s_top(double mu, alpha_s_options opt){
    auto alpha_s = QCD::strong_coupling{
      {opt.alpha_s, opt.mu0, nl_top}
    };
    return alpha_s(mu);
  }

  double alpha_s_bottom(double mu, double alpha_s_mZ){
    return alpha_s_bottom(mu, {mZ, alpha_s_mZ});
  }

  double alpha_s_bottom(double mu, alpha_s_options opt){
    auto alpha_s = QCD::strong_coupling{
      {
        opt.alpha_s, opt.mu0,
        (opt.mu0 < mu_thr)?nl_bottom:(nl_bottom + 1)
      }
    };
    if(opt.mu0 > mu_thr) alpha_s = alpha_s.decouple_down(
        {mu_thr, mb_SI, QCD::scheme::SI}
    );
    assert(alpha_s.nf() == nl_bottom);
    return alpha_s(mu);
  }

}
