#ifdef NDEBUG
#undef NDEBUG
#endif
#include <iostream>
#include <cmath>

#include "load_grid.hpp"
#include "xsection.hpp"
#include "threshold.hpp"
#include "mass_schemes.hpp"

int main(){
  using namespace QQbar_threshold::pert_order;
  using QQbar_threshold::production_channel;
  using QQbar_threshold::SM;
  using QQbar_threshold::ttbar_R_ratio;
  using QQbar_threshold::PSshift;
  using QQbar_threshold::load_grid;
  using QQbar_threshold::grid_directory;
  load_grid(grid_directory() + "ttbar_grid.tsv");
  auto opt = QQbar_threshold::top_options();

  constexpr double ep = 1e-12;

  constexpr double sqrt_s = 344;
  constexpr double mu = 80;
  constexpr double mt = 171.5;
  constexpr double width = 1.33;
  constexpr double mu_weak = 350.;

  opt.alpha = 1/137.03599907430635;
  opt.production = production_channel::photon_only;
  opt.beyond_QCD = SM::none;
  opt.expand_s = true;
  opt.mass_scheme = {PSshift,20};
  constexpr std::array<double, 2> results_photon = {{
      0.006086654273699477,
      0.004564520457571497
    }};
  for(order i = N2LO; i <= N3LO; ++i){
    double diff = std::abs(
        ttbar_R_ratio(sqrt_s, {mu, mu_weak}, {mt, width}, i, opt)
        - ttbar_R_ratio(sqrt_s, mu, {mt, width}, i, opt)
        - results_photon[i-2]
    );
    std::cout << i << ": " << diff;
    assert(diff < ep);
    std::cout << " (ok)\n";
  }
}
